from ntt_api_management import *
from datetime import time, date, timedelta, datetime


class NTTAPIDailyLimit(NTTAPIBase):
    def __init__(
        self,
        key: str = "",
        number_requests: int = 0,
        request_limit: int = 100,
        last_request_time: datetime = datetime.now(),
    ) -> None:
        super().__init__(key, last_request_time)
        self.__number_requests = number_requests
        self.__request_limit = request_limit

    # just for testing purposes, not used in the real implementation
    @property
    def number_requests(self) -> int:
        return self.__number_requests

    @property
    def request_limit(self) -> int:
        return self.__request_limit

    @property
    def last_access(self) -> date:
        return self.last_request_time.date()

    @last_access.setter
    def last_access(self, value: date) -> None:
        self.last_request_time = datetime.combine(value, time.min)

    # end of testing purposes

    def _make_request(self, **kwargs) -> None:
        super()._make_request(**kwargs)
        self.__number_requests += 1

    def ToDict(self) -> dict[str, object]:
        baseDict = super().ToDict()
        baseDict["number_requests"] = self.__number_requests
        baseDict["request_limit"] = self.__request_limit

        return baseDict

    def _PreprocessIncomingDict(self, data: dict[str, object]) -> None:
        super()._PreprocessIncomingDict(data)

    def _FromDict(self, data: dict[str, object]) -> None:
        super()._FromDict(data)
        self.__number_requests = data["number_requests"]
        self.__request_limit = data["request_limit"]

    def can_be_used(self) -> bool:
        return self.__number_requests < self.request_limit

    def should_reset(self) -> bool:
        return self.last_access != date.today()

    def reset(self) -> None:
        self.__number_requests = 0
        self.last_access = date.today()

    def __repr__(self) -> str:
        return f"<NTT API: {self.key} - Requests: {self.number_requests} />"



from abc import ABCMeta
from datetime import datetime
from .i_ntt_api import *


class NTTAPIBase(INTTAPI):
    def __init__(
        self,
        key: str,
        last_request_time: datetime = datetime.now(),
    ) -> None:
        self.__key = key
        self.__last_request_time = last_request_time

    @property
    def key(self) -> str:
        return self.__key

    @property
    def last_request_time(self) -> datetime:
        return self.__last_request_time

    @last_request_time.setter
    def last_request_time(self, value: datetime) -> None:
        self.__last_request_time = value

    def make_request(self, **kwargs) -> None:
        self.__last_request_time = datetime.now()
        self._make_request(**kwargs)

    @abstractmethod
    def _make_request(self, **kwargs) -> None:
        pass

    def ToDict(self) -> dict[str, object]:
        return {
            "key": self.key,
            "last_request_time": self.__last_request_time.strftime("%Y-%m-%d %H:%M:%S"),
        }

    def FromDict(self, data: dict[str, object]) -> INTTAPI:
        self._PreprocessIncomingDict(data)
        self._FromDict(data)
        return self

    def _FromDict(self, data: dict[str, object]) -> None:
        self.__key = data["key"]
        self.__last_request_time = data["last_request_time"]

    def _PreprocessIncomingDict(self, data: dict[str, object]) -> None:
        data["last_request_time"] = datetime.strptime(
            data["last_request_time"], "%Y-%m-%d %H:%M:%S"
        )

    def should_reset(self) -> bool:
        return False

    def reset(self) -> None:
        pass

    def __repr__(self) -> str:
        return f"<NTT API: {self.key} />"



from datetime import date, datetime, timedelta
import json
import os
from typing import Optional, List
from .i_ntt_api import INTTAPI


class NTTAPIManagement:
    date_key = "date"
    data_key = "data"

    def __init__(
        self,
        api_model: INTTAPI,
        data_json_file: Optional[str] = None,
    ) -> None:
        self.__apis: List[INTTAPI] = []
        self.__current_api_index: int = -1
        self.__data_json_file = data_json_file
        self.__api_model = api_model

        # check data_json_file exists
        if data_json_file is not None and os.path.exists(data_json_file):
            self.__LoadDataFromFile()

    def __LoadDataFromFile(self) -> None:
        if self.__data_json_file is None:
            return

        try:
            with open(self.__data_json_file, "r") as file:
                data_dicts = json.loads(file.read())
        except:
            self.UpdateDataFile()
            return

        self.__apis = [self.__api_model().FromDict(data) for data in data_dicts]

        self.__CheckResetAPIs()
        self.__UpdateAPIIndex()

    def __CheckResetAPIs(self) -> None:
        for api in self.__apis:
            if api.should_reset():
                api.reset()

        self.UpdateDataFile()

    def UpdateDataFile(self) -> None:
        if self.__data_json_file is None:
            return

        data_dicts = [api.ToDict() for api in self.__apis]

        with open(self.__data_json_file, "w") as file:
            file.write(json.dumps(data_dicts))

    @property
    def API(self) -> INTTAPI:
        self.__UpdateAPIIndex()

        if self.__current_api_index < 0:
            return None

        return self.__apis[self.__current_api_index]

    @property
    def APIs(self) -> List[INTTAPI]:
        return self.__apis

    def add_api(self, api_key: str, **kwargs) -> None:
        self.__apis.append(self.__api_model(key=api_key, **kwargs))

        self.UpdateDataFile()
        self.__UpdateAPIIndex()

    def make_request(self) -> None:
        if self.API is None:
            return

        self.API.make_request()

        self.__UpdateAPIIndex()
        self.UpdateDataFile()

    def __UpdateAPIIndex(self) -> None:
        if len(self.__apis) == 0:
            pass
        else:
            for i in range(len(self.__apis)):
                if self.__apis[i].can_be_used():
                    self.__current_api_index = i
                    return

        self.__current_api_index = -1



import os
import unittest
from ntt_api_management import *
from .ntt_minute_limit_api import *


class APIManagementMinuteLimitTest(unittest.TestCase):
    test_data_file = "test_minute_limit_data.json"
    first_api = "first-api"
    second_api = "second"

    def setUp(self) -> None:
        self.__CleanDataFile()

        self.apiWith2APIs = NTTAPIManagement(
            data_json_file=self.test_data_file,
            api_model=NTTAPIMinuteLimit,
        )
        self.apiWith2APIs.add_api(self.first_api)
        self.apiWith2APIs.add_api(self.second_api)

    def tearDown(self) -> None:
        self.__CleanDataFile()

    def __CleanDataFile(self) -> None:
        if os.path.exists(self.test_data_file):
            os.remove(self.test_data_file)

    def test_GivenAPIs_WhenMakeRequest_ThenTheAPIIsChanged(self) -> None:
        api = NTTAPIManagement(api_model=NTTAPIMinuteLimit)

        api.add_api(self.first_api)
        api.add_api(self.second_api)

        api.make_request()

        self.assertEqual(api.API.key, self.second_api)

    def test_GivenAPI_WhenMakeRequestThatTheFirstOneRequestLessThan1Minute_ThenAPICannotBeCalled(
        self,
    ) -> None:
        for _ in range(2):
            self.apiWith2APIs.make_request()

        self.assertIsNone(self.apiWith2APIs.API)

    def test_GivenAPI_WhenCreateTheNewInstance_ThenTheAPIsAreShared(
        self,
    ) -> None:
        self.apiWith2APIs.make_request()

        other_api = NTTAPIManagement(
            NTTAPIMinuteLimit,
            data_json_file=self.test_data_file,
        )

        self.assertEqual(other_api.API.key, self.second_api)

    def test_GivenAPIWithPrevRequestIs2MinutesAgo_WhenAskAPI_ThenItShouldReturn(
        self,
    ) -> None:
        self.apiWith2APIs.make_request()
        self.apiWith2APIs.make_request()
        self.__FakeDateTo2MinutesAgo(self.apiWith2APIs)

        self.assertEqual(self.apiWith2APIs.API.key, self.first_api)

    def __FakeDateTo2MinutesAgo(self, api: NTTAPIManagement) -> None:
        api.APIs[0].last_request_time = api.APIs[0].last_request_time - timedelta(
            minutes=2
        )



from abc import ABC, abstractmethod
from datetime import datetime


class INTTAPI(ABC):
    @abstractmethod
    def ToDict(self) -> dict[str, object]:
        pass

    @abstractmethod
    def FromDict(self, data: dict[str, object]) -> "INTTAPI":
        pass

    @abstractmethod
    def make_request(self, **kwargs) -> None:
        pass

    @abstractmethod
    def can_be_used(self) -> bool:
        pass

    @abstractmethod
    def should_reset(self) -> bool:
        pass

    @abstractmethod
    def reset(self) -> None:
        pass

    @property
    @abstractmethod
    def key(self) -> str:
        pass

    @property
    @abstractmethod
    def last_request_time(self) -> datetime:
        pass

    @last_request_time.setter
    @abstractmethod
    def last_request_time(self, value: datetime) -> None:
        pass



from ntt_api_management import *
from datetime import datetime, timedelta


class NTTAPIMinuteLimit(NTTAPIBase):
    def __init__(
        self,
        key: str = "",
        request_times: int = 0,
        last_request_time: datetime = datetime.now(),
    ) -> None:
        super().__init__(key, last_request_time)
        self.__request_times = request_times

    @property
    def number_requests(self) -> int:
        return self.__request_times

    def _make_request(self, **kwargs) -> None:
        super()._make_request(**kwargs)
        self.__request_times += 1

    def can_be_used(self) -> bool:
        if self.__request_times == 0:
            return True
        elif self.last_request_time < datetime.now() - timedelta(minutes=1):
            return True
        else:
            return False

    def ToDict(self) -> dict[str, object]:
        baseDict = super().ToDict()

        baseDict["request_times"] = self.__request_times
        return baseDict

    def _PreprocessIncomingDict(self, data: dict[str, object]) -> None:
        super()._PreprocessIncomingDict(data)

    def _FromDict(self, data: dict[str, object]) -> None:
        super()._FromDict(data)
        self.__request_times = data["request_times"]



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    description = f.read()

setup(
    name="ntt-api-management",
    version="1.0.0",
    author="threezinedine",
    author_email="threezinedine@gmail.com",
    description=description,
    packages=find_packages(),
    install_requires=[],
    keywords=["API", "Management", "ntt"],
)



from datetime import date, timedelta
import json
import os
from parameterized import parameterized
from typing import List
import unittest
from ntt_api_management import *
from .ntt_daily_limit_api import *


class APIManagementDailyLimitTest(unittest.TestCase):
    test_data_file = "test_data.json"
    first_api = "first-api"
    second_api = "second-api"
    test_limit = 10

    def setUp(self) -> None:
        self.__CleanDataFile()
        NTTAPIDailyLimit.daily_limit = self.test_limit

        self.apiWith2APIs = NTTAPIManagement(
            data_json_file=self.test_data_file,
            api_model=NTTAPIDailyLimit,
        )
        self.apiWith2APIs.add_api(self.first_api, request_limit=self.test_limit)
        self.apiWith2APIs.add_api(self.second_api, request_limit=self.test_limit)

    def tearDown(self) -> None:
        self.__CleanDataFile()

    def __CleanDataFile(self) -> None:
        if os.path.exists(self.test_data_file):
            os.remove(self.test_data_file)

    def test_GivenAPIManagement_WhenAskAPI_ThenReturnEmpty(self) -> None:
        api = NTTAPIManagement(NTTAPIDailyLimit)

        self.assertIsNone(api.API)

    def test_GivenAPIManagementWithAddedAPI_WhenAskAPI_ThenReturnAPI(self) -> None:
        api = NTTAPIManagement(NTTAPIDailyLimit)

        api.add_api(self.first_api, request_limit=self.test_limit)

        self.assertEqual(api.API.key, self.first_api)
        self.__APIsShouldHaveKeys(api.APIs, [self.first_api])

    def __APIShouldHasKey(self, api: INTTAPI, key: str) -> None:
        self.assertEqual(api.key, key)

    def __APIsShouldHaveKeys(self, apis: List[INTTAPI], keys: List[str]) -> None:
        self.assertEqual(len(apis), len(keys))
        for i in range(len(apis)):
            self.__APIShouldHasKey(apis[i], keys[i])

    def test_GivenAPIManagementWithAddedAPI_WhenAddNewAPI_ThenThePreviousAPIIsUsed(
        self,
    ) -> None:
        api = NTTAPIManagement(NTTAPIDailyLimit)
        api.add_api(self.first_api, request_limit=self.test_limit)

        api.add_api(self.second_api, request_limit=self.test_limit)

        self.__APIShouldHasKey(api.API, self.first_api)
        self.__APIsShouldHaveKeys(api.APIs, [self.first_api, self.second_api])
        self.assertEqual(api.API.number_requests, 0)

    def test_GivenAPIManagementWithAddedAPI_WhenCreateOtherInstance_ThenTheAPIsAreShared(
        self,
    ) -> None:
        api = NTTAPIManagement(
            NTTAPIDailyLimit,
            data_json_file=self.test_data_file,
        )
        api.add_api(self.first_api, request_limit=self.test_limit)

        other_api = NTTAPIManagement(
            NTTAPIDailyLimit,
            data_json_file=self.test_data_file,
        )

        self.assertEqual(other_api.API.key, self.first_api)

    def test_GivenAddedAPI_WhenMakeARequest_ThenTheNumberOfRequestsIsIncreased(
        self,
    ) -> None:
        api = NTTAPIManagement(NTTAPIDailyLimit)
        api.add_api(self.first_api, request_limit=self.test_limit)

        api.make_request()

        self.assertEqual(api.API.number_requests, 1)

    def test_GivenNoAPIAdded_WhenMakeARequest_ThenNoError(self) -> None:
        api = NTTAPIManagement(NTTAPIDailyLimit)

        api.make_request()

    def test_Given2APIsWhenTheFirstOneReachesTheLimit_ThenTheSecondOneIsUsed(
        self,
    ) -> None:
        limit = 10
        api = NTTAPIManagement(NTTAPIDailyLimit)
        first_api = "first-api"
        second_api = "second-api"
        api.add_api(first_api, request_limit=limit)
        api.add_api(second_api, request_limit=limit)

        for _ in range(limit):
            api.make_request()

        self.assertEqual(api.API.key, second_api)

    @parameterized.expand(
        [
            (5, 5, 0),
            (10, 10, 0),
            (11, 10, 1),
            (14, 10, 4),
        ]
    )
    def test_Given2APIs_ThenTheirNumberOfRequestsAreSave(
        self,
        make_requests,
        first_api_requests,
        second_api_requests,
    ) -> None:
        for _ in range(make_requests):
            self.apiWith2APIs.make_request()

        other_api = NTTAPIManagement(
            NTTAPIDailyLimit, data_json_file=self.test_data_file
        )

        self.assertEqual(len(other_api.APIs), len(self.apiWith2APIs.APIs))
        self.assertEqual(other_api.APIs[0].number_requests, first_api_requests)
        self.assertEqual(other_api.APIs[1].number_requests, second_api_requests)
        self.__ItShouldHasTheNumberOfRequestsAsExpected(
            other_api.APIs,
            [
                first_api_requests,
                second_api_requests,
            ],
        )

    def __ItShouldHasTheNumberOfRequestsAsExpected(
        self,
        apis: List[INTTAPI],
        requests: List[int],
    ) -> None:
        for i in range(len(apis)):
            self.assertEqual(apis[i].number_requests, requests[i])

    @parameterized.expand(
        [
            (5, 6, 0),
            (10, 10, 1),
            (11, 10, 2),
            (14, 10, 5),
            (18, 10, 9),
            (22, 10, 10),
        ]
    )
    def test_Given2APIs_WhenAnotherAPIMakeRequest_ThenTheNumberOfRequestsAreIncreased(
        self,
        previous_requests,
        first_api_requests,
        second_api_requests,
    ) -> None:
        for _ in range(previous_requests):
            self.apiWith2APIs.make_request()

        other_api = NTTAPIManagement(
            NTTAPIDailyLimit,
            data_json_file=self.test_data_file,
        )

        other_api.make_request()

        self.__ItShouldHasTheNumberOfRequestsAsExpected(
            other_api.APIs,
            [
                first_api_requests,
                second_api_requests,
            ],
        )

    def test_Given2APIs_WhenLoadTheFileWithError_ThenTheFileIsResetToDefaultValidDataFile(
        self,
    ) -> None:
        self.__FakeTheDataFileToInvalidJsonFormat()
        api = NTTAPIManagement(NTTAPIDailyLimit, data_json_file=self.test_data_file)

        self.assertIsNone(api.API)
        self.__APIsShouldHaveKeys(api.APIs, [])

    def __FakeTheDataFileToInvalidJsonFormat(self) -> None:
        with open(self.test_data_file, "w") as file:
            file.write('{"date": "2021-10-10"}, "data": [{"key"}')

    def test_Given2APIs_WhenLoadTheFirstTimeInToday_ThenTheNumberOfRequestsAreReset(
        self,
    ) -> None:
        for _ in range(5):
            self.apiWith2APIs.make_request()

        self.__FakeDateToYesterday(self.apiWith2APIs)

        other_api = NTTAPIManagement(
            NTTAPIDailyLimit, data_json_file=self.test_data_file
        )

        self.__ItShouldHasTheNumberOfRequestsAsExpected(
            other_api.APIs,
            [
                0,
                0,
            ],
        )

    def __FakeDateToYesterday(self, manager: NTTAPIManagement) -> None:
        for api in manager.APIs:
            api.last_access = date.today() - timedelta(days=1)

        manager.UpdateDataFile()
